import pathlib
import contextlib
import datetime
import os
import io
import json


from tasks.models import Task, Annotation, Prediction
from projects.models import Project
from data_import.models import FileUpload
from core.utils.io import get_data_dir
from data_manager.models import View, FilterGroup, Filter
from django.core.files.base import File
from io_storages.gcs.models import GCSImportStorage, GCSExportStorage
from io_storages.azure_blob.models import AzureBlobImportStorage, AzureBlobExportStorage
from io_storages.s3.models import S3ImportStorage, S3ExportStorage
from io_storages.redis.models import RedisImportStorage, RedisExportStorage
from ml.models import MLBackend
from core.utils.params import get_env


@contextlib.contextmanager
def suppress_autotime(model, fields):
    """allow to keep original created_at value for auto_now_add=True field"""
    _original_values = {}
    for field in model._meta.local_fields:
        if field.name in fields:
            _original_values[field.name] = {
                "auto_now": field.auto_now,
                "auto_now_add": field.auto_now_add,
            }
            field.auto_now = False
            field.auto_now_add = False
    try:
        yield
    finally:
        for field in model._meta.local_fields:
            if field.name in fields:
                field.auto_now = _original_values[field.name]["auto_now"]
                field.auto_now_add = _original_values[field.name]["auto_now_add"]


def _migrate_tasks(project_path, project):
    """Migrate tasks from json file to database objects"""
    tasks_path = project_path / "tasks.json"
    with io.open(os.path.abspath(tasks_path), encoding="utf-8") as t:
        tasks_data = json.load(t)
        for task_id, task_data in tasks_data.items():
            task = Task.objects.create(data=task_data.get("data", {}), project=project)

            # migrate annotations
            annotations_path = project_path / "completions" / "{}.json".format(task_id)
            if annotations_path.exists():
                with io.open(os.path.abspath(annotations_path), encoding="utf-8") as c:
                    annotations_data = json.load(c)
                    for annotation in annotations_data["completions"]:
                        task_annotation = Annotation(
                            result=annotation["result"],
                            task=task,
                            lead_time=annotation["lead_time"],
                            was_cancelled=annotation.get("was_cancelled", False),
                            completed_by=project.created_by,
                        )
                        with suppress_autotime(task_annotation, ["created_at"]):
                            task_annotation.created_at = datetime.datetime.fromtimestamp(
                                annotation["created_at"],
                                tz=datetime.datetime.now().astimezone().tzinfo,
                            )
                            task_annotation.save()

            # migrate predictions
            predictions_data = task_data.get("predictions", [])
            for prediction in predictions_data:
                task_prediction = Prediction(
                    result=prediction["result"], task=task, score=prediction.get("score")
                )
                with suppress_autotime(task_prediction, ["created_at"]):
                    task_prediction.created_at = datetime.datetime.fromtimestamp(
                        prediction["created_at"], tz=datetime.datetime.now().astimezone().tzinfo
                    )
                    task_prediction.save()


def _migrate_tabs(project_path, project):
    """Migrate tabs from tabs.json to Views table"""
    tabs_path = project_path / "tabs.json"
    if tabs_path.exists():
        with io.open(os.path.abspath(tabs_path), encoding="utf-8") as t:
            tabs_data = json.load(t)
            for tab in tabs_data["tabs"]:
                view = View.objects.create(project=project)
                tab["id"] = view.id
                ordering = tab.pop("ordering", None)
                selected_items = tab.pop("selectedItems", None)

                # migrate filters
                filter_group = None
                filters = tab.pop("filters", None)
                if filters is not None:
                    filter_group = FilterGroup.objects.create(
                        conjunction=filters.get("conjunction", "and")
                    )
                    if "items" in filters:
                        for f in filters["items"]:
                            view_filter = Filter.objects.create(
                                **{
                                    "column": f.get("filter", ""),
                                    "operator": f.get("operator", ""),
                                    "type": f.get("type", ""),
                                    "value": f.get("value", {}),
                                }
                            )
                            filter_group.filters.add(view_filter)
                hidden_columns = {"explore": [], "labeling": []}
                hidden_columns_data = tab.pop("hiddenColumns", None)

                # apply naming change to tabs internal data
                if hidden_columns_data is not None:
                    for c in hidden_columns_data.get("explore", []):
                        hidden_columns["explore"].append(c.replace("completion", "annotation"))
                    for c in hidden_columns_data.get("labeling", []):
                        hidden_columns["labeling"].append(c.replace("completion", "annotation"))
                    tab["hiddenColumns"] = hidden_columns
                view.data = tab
                view.ordering = ordering
                view.selected_items = selected_items
                view.filter_group = filter_group
                view.save()


def _migrate_storages(project, config):
    """Migrate source and target storages from config.json to database"""

    # source storages migration
    source = config.get("source", None)
    if source:
        if source.get("type") == "gcs":
            params = source.get("params", {})
            GCSImportStorage.objects.create(
                project=project,
                bucket=source.get("path"),
                prefix=params.get("prefix"),
                regex_filter=params.get("regex"),
                use_blob_urls=params.get("use_blob_urls"),
            )
        elif source.get("type") == "azure-blob":
            params = source.get("params", {})
            AzureBlobImportStorage.objects.create(
                project=project,
                container=source.get("path"),
                prefix=params.get("prefix"),
                regex_filter=params.get("regex"),
                use_blob_urls=params.get("use_blob_urls"),
            )
        elif source.get("type") == "s3":
            params = source.get("params", {})
            S3ImportStorage.objects.create(
                project=project,
                bucket=source.get("path"),
                prefix=params.get("prefix"),
                regex_filter=params.get("regex"),
                use_blob_urls=params.get("use_blob_urls"),
                region_name=params.get("region"),
            )
        elif source.get("type") == "redis":
            params = source.get("params", {})
            RedisImportStorage.objects.create(
                project=project,
                path=source.get("path"),
                host=params.get("host"),
                port=params.get("port"),
                password=params.get("password"),
                db=params.get("db", 1),
            )
    # target storages migration
    target = config.get("target", None)
    if target:
        if target.get("type") == "gcs":
            params = target.get("params", {})
            GCSExportStorage.objects.create(
                project=project,
                bucket=target.get("path"),
                prefix=params.get("prefix"),
                regex_filter=params.get("regex"),
                use_blob_urls=params.get("use_blob_urls"),
            )
        elif target.get("type") == "azure-blob":
            params = target.get("params", {})
            AzureBlobExportStorage.objects.create(
                project=project,
                container=target.get("path"),
                prefix=params.get("prefix"),
                regex_filter=params.get("regex"),
                use_blob_urls=params.get("use_blob_urls"),
            )
        elif target.get("type") == "s3":
            params = target.get("params", {})
            S3ExportStorage.objects.create(
                project=project,
                bucket=target.get("path"),
                prefix=params.get("prefix"),
                regex_filter=params.get("regex"),
                use_blob_urls=params.get("use_blob_urls"),
                region_name=params.get("region"),
            )
        elif target.get("type") == "redis":
            params = target.get("params", {})
            RedisExportStorage.objects.create(
                project=project,
                path=target.get("path"),
                host=params.get("host"),
                port=params.get("port"),
                password=params.get("password"),
                db=params.get("db", 1),
            )


def _migrate_ml_backends(project, config):
    """Migrate ml backend settings from config.json to database"""
    ml_backends = config.get("ml_backends", [])
    for ml_backend in ml_backends:
        MLBackend.objects.create(
            project=project, url=ml_backend.get("url"), title=ml_backend.get("name")
        )


def _migrate_uploaded_files(project, project_path):
    """Migrate files uploaded by user"""
    source_upload_path = project_path / "upload"
    if not source_upload_path.exists():
        return
    target_upload_path = (
        pathlib.Path(get_env("LABEL_STUDIO_BASE_DATA_DIR", get_data_dir())) / "upload"
    )
    if not target_upload_path.exists():
        os.makedirs(str(target_upload_path), exist_ok=True)

    src_files = os.listdir(str(source_upload_path))
    for file_name in src_files:
        full_file_name = os.path.join(str(source_upload_path), file_name)
        with open(full_file_name, "rb") as f:
            FileUpload.objects.create(
                user=project.created_by, project=project, file=File(f, name=file_name)
            )


def migrate_existing_project(project_path, project, config):
    """Migration projects from previous version of Label Studio"""

    _migrate_tasks(project_path, project)
    _migrate_tabs(project_path, project)
    _migrate_storages(project, config)
    _migrate_ml_backends(project, config)
    _migrate_uploaded_files(project, project_path)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import os

from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from io_storages.serializers import ImportStorageSerializer, ExportStorageSerializer
from io_storages.azure_blob.models import AzureBlobImportStorage, AzureBlobExportStorage


class AzureBlobImportStorageSerializer(ImportStorageSerializer):
    type = serializers.ReadOnlyField(default="azure")
    presign = serializers.BooleanField(required=False, default=True)

    class Meta:
        model = AzureBlobImportStorage
        fields = "__all__"

    def validate(self, data):
        data = super(AzureBlobImportStorageSerializer, self).validate(data)
        storage = AzureBlobImportStorage(**data)
        try:
            storage.validate_connection()
        except Exception as exc:
            raise ValidationError(exc)
        return data


class AzureBlobExportStorageSerializer(ExportStorageSerializer):
    type = serializers.ReadOnlyField(default="azure")

    class Meta:
        model = AzureBlobExportStorage
        fields = "__all__"



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import setuptools
import label_studio_withoutsignin
import sys

# python 3.9 warning
if sys.version_info[0] == 3 and sys.version_info[1] >= 9:
    print()
    print('~========================================================~')
    print('| Your python version is %i.%i and Label Studio has a      |' % (sys.version_info[0], sys.version_info[1]))
    print('| lot of installation problems with this Python version. |')
    print('| Please use Python 3.6 - 3.8 to avoid problems with     |')
    print('| your installation.                                     |')
    print('~========================================================~')
    print()

print(label_studio_withoutsignin.package_name, label_studio_withoutsignin.__version__)

# Readme
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

# Module dependencies
requirements, dependency_links = [], []
with open('deploy/requirements.txt') as f:
    for line in f.read().splitlines():
        if line.startswith('-e git+'):
            dependency_links.append(line.replace('-e ', ''))
        else:
            requirements.append(line)

setuptools.setup(
    name=label_studio_withoutsignin.package_name,
    version=label_studio_withoutsignin.__version__,
    author='dmytro.bershov',
    author_email="",
    description='Label Studio without sign in user access',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/DimaVinnitsa/label-studio',
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    install_requires=requirements,
    dependency_links=dependency_links,
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'label-studio=label_studio_withoutsignin.server:main',
        ],
    }
)



from django.db import transaction

from organizations.models import Organization, OrganizationMember


def create_organization(title, created_by):
    with transaction.atomic():
        org = Organization.objects.create(title=title, created_by=created_by)
        OrganizationMember.objects.create(user=created_by, organization=org)
        return org



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import pytest
import requests_mock
import json

from projects.models import Project
from ml.models import MLBackend
from tasks.models import Task, Prediction, Annotation
from tests.utils import make_project
from core.redis import redis_healthcheck

_project_for_text_choices_onto_A_B_classes = dict(
    title="Test",
    label_config="""
        <View>
          <Text name="meta_info" value="$meta_info"></Text>
          <Text name="text" value="$text"></Text>
          <Choices name="text_class" toName="text" choice="single">
            <Choice value="class_A"></Choice>
            <Choice value="class_B"></Choice>
          </Choices>
        </View>""",
)

_2_tasks_with_textA_and_textB = [
    {"meta_info": "meta info A", "text": "text A"},
    {"meta_info": "meta info B", "text": "text B"},
]

_2_prediction_results_for_textA_textB = [
    {
        "result": [
            {
                "from_name": "text_class",
                "to_name": "text",
                "type": "labels",
                "value": {"labels": ["class_A"], "start": 0, "end": 1},
            }
        ],
        "score": 0.95,
    },
    {
        "result": [
            {
                "from_name": "text_class",
                "to_name": "text",
                "type": "labels",
                "value": {"labels": ["class_B"], "start": 0, "end": 1},
            }
        ],
        "score": 0.59,
    },
]


def run_task_predictions(client, project, mocker):
    class TestJob:
        def __init__(self, job_id):
            self.id = job_id

    m = MLBackend.objects.filter(project=project.id).filter(url="http://localhost:8999").first()
    return client.post(f"/api/ml/{m.id}/predict")


@pytest.mark.skipif(
    not redis_healthcheck(), reason="Starting predictions requires Redis server enabled"
)
@pytest.mark.parametrize(
    "project_config, tasks, annotations, prediction_results, log_messages, model_version_in_request, use_ground_truth",
    [
        (
            # project config
            _project_for_text_choices_onto_A_B_classes,
            # tasks
            _2_tasks_with_textA_and_textB,
            # annotations
            [
                dict(
                    result=[
                        {
                            "from_name": "text_class",
                            "to_name": "text",
                            "type": "labels",
                            "value": {"labels": ["class_A"], "start": 0, "end": 1},
                        }
                    ],
                    ground_truth=True,
                ),
                dict(
                    result=[
                        {
                            "from_name": "text_class",
                            "to_name": "text",
                            "type": "labels",
                            "value": {"labels": ["class_B"], "start": 0, "end": 1},
                        }
                    ],
                    ground_truth=True,
                ),
            ],
            # prediction results
            _2_prediction_results_for_textA_textB,
            # log messages
            None,
            # model version in request
            "12345",
            False,
        ),
        (
            # project config
            _project_for_text_choices_onto_A_B_classes,
            # tasks
            _2_tasks_with_textA_and_textB,
            # annotations
            [
                dict(
                    result=[
                        {
                            "from_name": "text_class",
                            "to_name": "text",
                            "type": "labels",
                            "value": {"labels": ["class_A"], "start": 0, "end": 1},
                        }
                    ],
                    ground_truth=True,
                ),
                dict(
                    result=[
                        {
                            "from_name": "text_class",
                            "to_name": "text",
                            "type": "labels",
                            "value": {"labels": ["class_B"], "start": 0, "end": 1},
                        }
                    ],
                    ground_truth=True,
                ),
            ],
            # prediction results
            _2_prediction_results_for_textA_textB,
            # log messages
            None,
            # model version in request
            "12345",
            True,
        ),
    ],
)
@pytest.mark.django_db
def test_predictions(
    business_client,
    project_config,
    tasks,
    annotations,
    prediction_results,
    log_messages,
    model_version_in_request,
    use_ground_truth,
    mocker,
):

    # create project with predefined task set
    project = make_project(project_config, business_client.user)

    for task, annotation in zip(tasks, annotations):
        t = Task.objects.create(data=task, project=project)
        if use_ground_truth:
            Annotation.objects.create(task=t, **annotation)

    # run prediction
    with requests_mock.Mocker() as m:
        m.post(
            "http://localhost:8999/setup",
            text=json.dumps({"model_version": model_version_in_request}),
        )
        m.post(
            "http://localhost:8999/predict",
            text=json.dumps(
                {"results": prediction_results[:1], "model_version": model_version_in_request}
            ),
        )
        r = run_task_predictions(business_client, project, mocker)
        assert r.status_code == 200
        assert m.called

    # check whether stats are created
    predictions = Prediction.objects.all()
    project = Project.objects.get(id=project.id)
    ml_backend = MLBackend.objects.get(url="http://localhost:8999")

    assert predictions.count() == len(tasks)

    for actual_prediction, expected_prediction_result in zip(predictions, prediction_results):
        assert actual_prediction.result == prediction_results[0]["result"]
        assert actual_prediction.score == prediction_results[0]["score"]
        assert ml_backend.model_version == actual_prediction.model_version


@pytest.mark.skipif(
    not redis_healthcheck(), reason="Starting predictions requires Redis server enabled"
)
@pytest.mark.parametrize(
    "test_name, project_config, setup_returns_model_version, tasks, annotations, "
    "input_predictions, prediction_call_count, num_project_stats, num_ground_truth_in_stats, "
    "num_ground_truth_fit_predictions",
    [
        (
            # test name just for reference
            "All predictions are outdated, project.model_version is outdated too",
            # project config: contains old model version
            dict(
                title="Test",
                model_version="12345_old",
                label_config="""
                <View>
                  <Text name="txt" value="$text"></Text>
                  <Choices name="cls" toName="txt" choice="single">
                    <Choice value="class_A"></Choice>
                    <Choice value="class_B"></Choice>
                  </Choices>
                </View>""",
            ),
            # setup API returns this model version
            "12345",
            # task data
            [{"text": "text A"}, {"text": "text B"}],
            # annotations: there is no any annotations
            [None, None],
            # predictions: 2 predictions are from old model version
            [
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_A"]},
                        }
                    ],
                    "score": 0.95,
                    "model_version": "12345_old",
                },
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_B"]},
                        }
                    ],
                    "score": 0.59,
                    "model_version": "12345_old",
                },
            ],
            # prediction call count is 2 for both tasks with old predictions
            2,
            # ground_truth stats
            0,
            0,
            0,
        ),
        (
            # test name just for reference
            "All predictions are up-to-date",
            # project config: contains actual model version
            dict(
                title="Test",
                model_version="12345_old",
                label_config="""
        <View>
          <Text name="txt" value="$text"></Text>
          <Choices name="cls" toName="txt" choice="single">
            <Choice value="class_A"></Choice>
            <Choice value="class_B"></Choice>
          </Choices>
        </View>""",
            ),
            # setup API returns this model version
            "12345",
            # task data
            [{"text": "text A"}, {"text": "text B"}],
            # annotations: there is no any annotations
            [None, None],
            # predictions: 2 predictions are from old model version
            [
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_A"]},
                        }
                    ],
                    "score": 0.95,
                    "model_version": "12345",
                },
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_B"]},
                        }
                    ],
                    "score": 0.59,
                    "model_version": "12345",
                },
            ],
            # prediction call count is 0 since predictions are up to date
            0,
            # ground_truth stats
            0,
            0,
            0,
        ),
        (
            # test name just for reference
            "Some predictions are outdated, other are up-to-date. project.model_version is up-to-date",
            # project config: contains actual model version
            dict(
                title="Test",
                model_version="12345",
                label_config="""
        <View>
          <Text name="txt" value="$text"></Text>
          <Choices name="cls" toName="txt" choice="single">
            <Choice value="class_A"></Choice>
            <Choice value="class_B"></Choice>
          </Choices>
        </View>""",
            ),
            # setup API returns this model version
            "12345",
            # task data
            [{"text": "text A"}, {"text": "text B"}],
            # annotations: there is no any annotations
            [None, None],
            # predictions: 2 predictions, one from the new model version, second from old
            [
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_A"]},
                        }
                    ],
                    "score": 0.95,
                    "model_version": "12345",
                },
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_B"]},
                        }
                    ],
                    "score": 0.59,
                    "model_version": "12345_old",
                },
            ],
            # prediction call count is 1 only for the task with old predictions
            1,
            # ground_truth stats
            0,
            0,
            0,
        ),
        (
            # test name just for reference
            "Some predictions are outdated, other are up-to-date. project.model_version is outdated",
            # project config: contains actual model version
            dict(
                title="Test",
                model_version="12345_old",
                label_config="""
<View>
  <Text name="txt" value="$text"></Text>
  <Choices name="cls" toName="txt" choice="single">
    <Choice value="class_A"></Choice>
    <Choice value="class_B"></Choice>
  </Choices>
</View>""",
            ),
            # setup API returns this model version
            "12345",
            # task data
            [{"text": "text A"}, {"text": "text B"}],
            # annotations: there is no any annotations
            [None, None],
            # predictions: 2 predictions, one from the new model version, second from old
            [
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_A"]},
                        }
                    ],
                    "score": 0.95,
                    "model_version": "12345",
                },
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_B"]},
                        }
                    ],
                    "score": 0.59,
                    "model_version": "12345_old",
                },
            ],
            # prediction call count is 1 only for the task with old predictions
            1,
            # ground_truth stats
            0,
            0,
            0,
        ),
        (
            # test name just for reference
            "All tasks has no predictions",
            # project config: contains actual model version
            dict(
                title="Test",
                model_version="12345",
                label_config="""
<View>
  <Text name="txt" value="$text"></Text>
  <Choices name="cls" toName="txt" choice="single">
    <Choice value="class_A"></Choice>
    <Choice value="class_B"></Choice>
  </Choices>
</View>""",
            ),
            # setup API returns this model version
            "12345",
            # task data
            [{"text": "text A"}, {"text": "text B"}],
            # annotations: there is no any annotations
            [None, None],
            # there is no any predictions yet
            [None, None],
            # prediction call count for all tasks without predictions
            2,
            # ground_truth stats
            0,
            0,
            0,
        ),
        (
            # test name just for reference
            "Some tasks has no predictions, others are up-to-date",
            # project config: contains actual model version
            dict(
                title="Test",
                model_version="12345",
                label_config="""
                <View>
                <Text name="txt" value="$text"></Text>
                <Choices name="cls" toName="txt" choice="single">
                <Choice value="class_A"></Choice>
                <Choice value="class_B"></Choice>
                </Choices>
                </View>""",
            ),
            # setup API returns this model version
            "12345",
            # task data
            [{"text": "text A"}, {"text": "text B"}],
            # annotations: there is no any annotations
            [None, None],
            # there is only one prediction (since job has finished before processing all tasks)
            [
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_A"]},
                        }
                    ],
                    "score": 0.95,
                    "model_version": "12345",
                },
                None,
            ],
            # prediction call count for all tasks without predictions
            1,
            # ground_truth stats
            0,
            0,
            0,
        ),
        (
            # test name just for reference
            "Some tasks has no predictions, others are up-to-date, labeled task contains ground_truth",
            # project config: contains actual model version
            dict(
                title="Test",
                model_version="12345",
                label_config="""
        <View>
        <Text name="txt" value="$text"></Text>
        <Choices name="cls" toName="txt" choice="single">
        <Choice value="class_A"></Choice>
        <Choice value="class_B"></Choice>
        </Choices>
        </View>""",
            ),
            # setup API returns this model version
            "12345",
            # task data
            [{"text": "text A"}, {"text": "text B"}],
            # annotations: first task has fitted ground_truth
            [
                None,
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_A"]},
                        }
                    ],
                    "ground_truth": True,
                },
            ],
            # there is only one prediction (since job has finished before processing all tasks)
            [
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_A"]},
                        }
                    ],
                    "score": 0.95,
                    "model_version": "12345",
                },
                None,
            ],
            # prediction call count for all tasks without predictions
            1,
            # ground_truth stats
            1,
            1,
            1,
        ),
        (
            # test name just for reference
            "Some tasks has no predictions, others are outdated",
            # project config: contains actual model version
            dict(
                title="Test",
                model_version="12345",
                label_config="""
        <View>
        <Text name="txt" value="$text"></Text>
        <Choices name="cls" toName="txt" choice="single">
        <Choice value="class_A"></Choice>
        <Choice value="class_B"></Choice>
        </Choices>
        </View>""",
            ),
            # setup API returns this model version
            "12345",
            # task data
            [{"text": "text A"}, {"text": "text B"}],
            # annotations: there is no any annotations
            [None, None],
            # there is only one prediction (since job has finished before processing all tasks)
            [
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_A"]},
                        }
                    ],
                    "score": 0.95,
                    "model_version": "12345_old",
                },
                None,
            ],
            # prediction call count for all tasks without up-to-date predictions
            2,
            # ground_truth stats
            0,
            0,
            0,
        ),
        (
            # test name just for reference
            "Some tasks has no predictions, others are outdated, project.model_version is outdated",
            # project config: contains actual model version
            dict(
                title="Test",
                model_version="12345_old",
                label_config="""
    <View>
    <Text name="txt" value="$text"></Text>
    <Choices name="cls" toName="txt" choice="single">
    <Choice value="class_A"></Choice>
    <Choice value="class_B"></Choice>
    </Choices>
    </View>""",
            ),
            # setup API returns this model version
            "12345",
            # task data
            [{"text": "text A"}, {"text": "text B"}],
            # annotations: there is no any annotations
            [None, None],
            # there is only one prediction (since job has finished before processing all tasks)
            [
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_A"]},
                        }
                    ],
                    "score": 0.95,
                    "model_version": "12345_old",
                },
                None,
            ],
            # prediction call count for all tasks without up-to-date predictions
            2,
            # ground_truth stats
            0,
            0,
            0,
        ),
        (
            # test name just for reference
            "Some tasks has no predictions, others are outdated, others are up-to-date",
            # project config: contains actual model version
            dict(
                title="Test",
                model_version="12345_old",
                label_config="""
<View>
<Text name="txt" value="$text"></Text>
<Choices name="cls" toName="txt" choice="single">
<Choice value="class_A"></Choice>
<Choice value="class_B"></Choice>
</Choices>
</View>""",
            ),
            # setup API returns this model version
            "12345",
            # task data
            [{"text": "text A"}, {"text": "text A"}, {"text": "text B"}],
            # annotations: there is no any annotations
            [None, None, None],
            # there is only one prediction (since job has finished before processing all tasks)
            [
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_A"]},
                        }
                    ],
                    "score": 0.95,
                    "model_version": "12345_old",
                },
                {
                    "result": [
                        {
                            "from_name": "cls",
                            "to_name": "txt",
                            "type": "choices",
                            "value": {"choices": ["class_A"]},
                        }
                    ],
                    "score": 0.95,
                    "model_version": "12345",
                },
                None,
            ],
            # prediction call count for all tasks without up-to-date predictions
            2,
            # ground_truth stats
            0,
            0,
            0,
        ),
    ],
)
@pytest.mark.django_db
def test_predictions_with_partially_predicted_tasks(
    business_client,
    test_name,
    setup_returns_model_version,
    project_config,
    tasks,
    annotations,
    input_predictions,
    prediction_call_count,
    num_project_stats,
    num_ground_truth_in_stats,
    num_ground_truth_fit_predictions,
    mocker,
):
    project = make_project(project_config, business_client.user)
    ml_backend = MLBackend.objects.get(url="http://localhost:8999")
    ml_backend.model_version = project_config["model_version"]
    ml_backend.save()
    for task, annotation, prediction in zip(tasks, annotations, input_predictions):
        task_obj = Task.objects.create(project=project, data=task)
        if annotation is not None:
            Annotation.objects.create(task=task_obj, **annotation)
        if prediction is not None:
            Prediction.objects.create(task=task_obj, **prediction)

    # run prediction
    with requests_mock.Mocker() as m:
        m.register_uri(
            "POST",
            "http://localhost:8999/setup",
            text=json.dumps({"model_version": setup_returns_model_version}),
        )  # noqa
        m.register_uri(
            "POST",
            "http://localhost:8999/predict",
            text=json.dumps(
                {
                    "results": [
                        {
                            "result": [
                                {
                                    "from_name": "cls",
                                    "to_name": "txt",
                                    "type": "choices",
                                    "value": {"choices": ["class_A"]},
                                }
                            ],
                            "score": 1,
                        }
                    ],
                    "model_version": setup_returns_model_version,
                }
            ),
        )

        r = run_task_predictions(business_client, project, mocker)
        assert r.status_code == 200
        assert (
            len(list(filter(lambda h: h.url.endswith("predict"), m.request_history)))
            == prediction_call_count
        )

        assert Prediction.objects.filter(
            task__project=project.id, model_version=setup_returns_model_version
        ).count() == len(tasks)
        assert (
            MLBackend.objects.get(url="http://localhost:8999").model_version
            == setup_returns_model_version
        )



# Generated by Django 3.1.4 on 2021-04-30 13:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("projects", "0008_auto_20210314_1840"),
    ]

    operations = [
        migrations.AddField(
            model_name="project",
            name="evaluate_predictions_automatically",
            field=models.BooleanField(
                default=False,
                help_text="Retrieve and display predictions when loading a task",
                verbose_name="evaluate predictions automatically",
            ),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import pytest
import json

from tests.utils import project_id


@pytest.mark.django_db
def test_signup_setting(business_client, client, settings):
    settings.DISABLE_SIGNUP_WITHOUT_LINK = True
    response = client.post(
        "/user/signup", data={"email": "test_user@example.com", "password": "test_password"}
    )
    assert response.status_code == 403

    response = business_client.get("/api/invite")

    invite_url = response.json()["invite_url"]

    response = client.post(
        invite_url, data={"email": "test_user@example.com", "password": "test_password"}
    )
    assert response.status_code == 302


@pytest.mark.django_db
def test_reset_token(business_client, client, settings):
    settings.DISABLE_SIGNUP_WITHOUT_LINK = True

    # get invite_url link and check it works
    response = business_client.get("/api/invite")
    invite_url = response.json()["invite_url"]
    response = client.post(
        invite_url, data={"email": "test_user@example.com", "password": "test_password"}
    )
    assert response.status_code == 302

    response = business_client.post("/api/invite/reset-token")
    new_invite_url = response.json()["invite_url"]

    # after reset old link returns permission denied
    client.logout()
    response = client.post(
        invite_url, data={"email": "test_user1@example.com", "password": "test_password"}
    )
    assert response.status_code == 403, response.content

    # but new one works fine
    response = client.post(
        new_invite_url, data={"email": "test_user2@example.com", "password": "test_password"}
    )
    assert response.status_code == 302



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from .azure_blob.models import (
    AzureBlobImportStorage,
    AzureBlobImportStorageLink,
    AzureBlobExportStorage,
    AzureBlobExportStorageLink,
)
from .s3.models import S3ImportStorage, S3ImportStorageLink, S3ExportStorage, S3ExportStorageLink
from .gcs.models import (
    GCSImportStorage,
    GCSImportStorageLink,
    GCSExportStorage,
    GCSExportStorageLink,
)
from .redis.models import (
    RedisImportStorage,
    RedisImportStorageLink,
    RedisExportStorage,
    RedisExportStorageLink,
)



# Generated by Django 3.1.4 on 2021-03-22 12:35

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("organizations", "0002_auto_20210310_2044"),
        ("users", "0002_auto_20210308_1559"),
    ]

    operations = [
        migrations.AddField(
            model_name="user",
            name="active_organization",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="active_users",
                to="organizations.organization",
            ),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import os
import io
import logging

from django.http import FileResponse
from django.conf import settings
from core.permissions import IsBusiness, view_with_auth

from .models import FileUpload
from ranged_fileresponse import RangedFileResponse

logger = logging.getLogger(__name__)


@view_with_auth(["GET"], (IsBusiness,))
def get_uploaded_file(request, filename):
    file = settings.UPLOAD_DIR + ("/" if not settings.UPLOAD_DIR.endswith("/") else "") + filename
    logger.debug(f"Fetch uploaded file by user {request.user} => {file}")
    file_upload = FileUpload.objects.get(file=file)

    return RangedFileResponse(request, open(file_upload.file.path, mode="rb"))



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django.urls import path, include

from . import api


app_name = "tasks"

_api_urlpatterns = [
    # CRUD
    path("", api.TaskListAPI.as_view(), name="task-list"),
    path("<int:pk>/", api.TaskAPI.as_view(), name="task-detail"),
    path("<int:pk>/annotations/", api.AnnotationsListAPI.as_view(), name="task-annotations"),
    path("<int:pk>/drafts", api.AnnotationDraftListAPI.as_view(), name="task-drafts"),
    path(
        "<int:pk>/annotations/<int:annotation_id>/drafts",
        api.AnnotationDraftListAPI.as_view(),
        name="task-annotations-drafts",
    ),
]

_api_annotations_urlpatterns = [
    path("<int:pk>/", api.AnnotationAPI.as_view(), name="annotation-detail"),
]

_api_drafts_urlpatterns = [
    path("<int:pk>/", api.AnnotationDraftAPI.as_view(), name="draft-detail"),
]

urlpatterns = [
    path("api/tasks/", include((_api_urlpatterns, app_name), namespace="api")),
    # TODO: these should be moved to the separate apps
    path(
        "api/annotations/",
        include((_api_annotations_urlpatterns, app_name), namespace="api-annotations"),
    ),
    path("api/drafts/", include((_api_drafts_urlpatterns, app_name), namespace="api-drafts")),
]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging

from copy import copy, deepcopy
from data_manager.functions import DataManagerException
from core.utils.common import timestamp_now

logger = logging.getLogger(__name__)


def propagate_annotations(project, queryset, **kwargs):
    items = queryset.value_list("id", flat=True)

    if len(items) < 2:
        raise DataManagerException(
            "Select more than two tasks, the first task annotation will be picked as source"
        )

    # check first annotation
    completed_task = items[0]
    task = project.target_storage.get(completed_task)
    if task is None or len(task.get("annotations", [])) == 0:
        raise DataManagerException(
            "The first selected task with ID = "
            + str(completed_task)
            + " should have at least one annotation to propagate"
        )

    # get first annotation
    source_annotation = task["annotations"][0]

    # copy first annotation to new annotations for each task
    for i in items[1:]:
        task = project.target_storage.get(i)
        if task is None:
            task = project.source_storage.get(i)
        annotation = deepcopy(source_annotation)

        # start annotation id from task_id * 9000
        annotations = task.get("annotations", None) or [{"id": i * 9000}]
        annotation["id"] = max([c["id"] for c in annotations]) + 1
        annotation["created_at"] = timestamp_now()

        if "annotations" not in task:
            task["annotations"] = []
        task["annotations"].append(annotation)

        project.target_storage.set(i, task)

    return {"response_code": 200}


def predictions_to_annotations(project, items, **kwargs):
    for i in items:
        task = project.source_storage.get(i)
        predictions = task.get("predictions", [])
        if len(predictions) == 0:
            continue

        prediction = predictions[-1]

        # load task with annotation from target storage
        task_with_annotations = project.target_storage.get(i)
        task = copy(task if task_with_annotations is None else task_with_annotations)

        annotations = task.get("annotations", None) or [{"id": i * 9000}]
        annotation = {
            "id": max([c["id"] for c in annotations]) + 1,
            "created_at": timestamp_now(),
            "lead_time": 0,
            "result": prediction.get("result", []),
        }

        if "annotations" not in task:
            task["annotations"] = []
        task["annotations"].append(annotation)

        project.target_storage.set(i, task)

    return {"response_code": 200}


actions = [
    {
        "entry_point": propagate_annotations,
        "title": "Propagate annotations",
        "order": 1,
        "experimental": True,
        "dialog": {
            "text": "This action will pick the first annotation from the first selected task, "
            "create new annotations for all selected tasks, "
            "and propagate the first annotation to others. "
            + "." * 80
            + "1. Create the first annotation for task A. "
            "2. Select task A with checkbox as first selected item. "
            "3. Select other tasks where you want to copy the first annotation from task A. "
            "4. Click Propagate annotations. "
            + "." * 80
            + "! Warning: it is an experimental feature! It could work well with Choices, "
            "but other annotation types (RectangleLabels, Text Labels, etc) "
            "will have a lot of issues.",
            "type": "confirm",
        },
    },
    {
        "entry_point": predictions_to_annotations,
        "title": "Predictions => annotations",
        "order": 1,
        "experimental": True,
        "dialog": {
            "text": "This action will create a new annotation from the last task prediction for each selected task.",
            "type": "confirm",
        },
    },
]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django.conf import settings
from rest_framework import serializers

from tasks.models import Task, Annotation
from tasks.serializers import PredictionSerializer
from users.models import User
from core.label_config import replace_task_data_undefined_with_config_field


class CompletedBySerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ["id", "email", "first_name", "last_name"]


class AnnotationSerializer(serializers.ModelSerializer):
    completed_by = CompletedBySerializer(read_only=True)

    class Meta:
        model = Annotation
        fields = "__all__"


class ExportDataSerializer(serializers.ModelSerializer):
    annotations = AnnotationSerializer(many=True, read_only=True)
    predictions = PredictionSerializer(many=True, read_only=True)
    file_upload = serializers.ReadOnlyField(source="file_upload_name")

    # resolve $undefined$ key in task data, if any
    def to_representation(self, task):
        project = task.project
        data = task.data

        replace_task_data_undefined_with_config_field(data, project)

        return super().to_representation(task)

    class Meta:
        model = Task
        exclude = ("overlap", "is_labeled")



# Generated by Django 3.1.4 on 2021-03-08 11:41

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("tasks", "0002_auto_20210304_1423"),
        ("tasks", "0002_auto_20210305_2035"),
    ]

    operations = []



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import drf_yasg.openapi as openapi
import logging
import numpy as np
import pathlib
import os

from collections import Counter
from django.db import IntegrityError
from django.db.models.fields import DecimalField
from django.conf import settings
from drf_yasg.utils import swagger_auto_schema
from django.db.models import Q, When, Count, Case, OuterRef, Max, Exists, Value, BooleanField
from rest_framework import generics, status, filters
from rest_framework.exceptions import NotFound, ValidationError as RestValidationError
from rest_framework.parsers import FormParser, JSONParser, MultiPartParser
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.views import exception_handler

from core.utils.common import conditional_atomic
from core.label_config import config_essential_data_has_changed
from projects.models import Project, ProjectSummary
from projects.serializers import (
    ProjectSerializer,
    ProjectLabelConfigSerializer,
    ProjectSummarySerializer,
)
from tasks.models import Task, Annotation, Prediction, TaskLock
from tasks.serializers import TaskSerializer, TaskWithAnnotationsAndPredictionsAndDraftsSerializer

from core.mixins import APIViewVirtualRedirectMixin, APIViewVirtualMethodMixin
from core.permissions import all_permissions, ViewClassPermission
from core.utils.common import (
    get_object_with_check_and_log,
    bool_from_request,
    paginator,
    paginator_help,
)
from core.utils.exceptions import ProjectExistException, LabelStudioDatabaseException
from core.utils.io import find_dir, find_file, read_yaml

from data_manager.functions import get_prepared_queryset
from data_manager.models import View

logger = logging.getLogger(__name__)


_result_schema = openapi.Schema(
    title="Labeling result",
    description="Labeling result (choices, labels, bounding boxes, etc.)",
    type=openapi.TYPE_OBJECT,
    properies={
        "from_name": openapi.Schema(
            title="from_name",
            description="The name of the labeling tag from the project config",
            type=openapi.TYPE_STRING,
        ),
        "to_name": openapi.Schema(
            title="to_name",
            description="The name of the labeling tag from the project config",
            type=openapi.TYPE_STRING,
        ),
        "value": openapi.Schema(
            title="value",
            description="Labeling result value. Format depends on chosen ML backend",
            type=openapi.TYPE_OBJECT,
        ),
    },
    example={"from_name": "image_class", "to_name": "image", "value": {"labels": ["Cat"]}},
)

_task_data_schema = openapi.Schema(
    title="Task data",
    description="Task data",
    type=openapi.TYPE_OBJECT,
    example={"id": 1, "my_image_url": "https://app.heartex.ai/static/samples/kittens.jpg"},
)


class ProjectListAPI(generics.ListCreateAPIView):
    """
    get:
    List your projects

    Return a list of the projects that you've created.

    post:
    Create new project

    Create a labeling project.
    """

    parser_classes = (JSONParser, FormParser, MultiPartParser)
    serializer_class = ProjectSerializer
    filter_backends = [filters.OrderingFilter]
    permission_required = ViewClassPermission(
        GET=all_permissions.projects_view,
        POST=all_permissions.projects_create,
    )
    ordering = ["-created_at"]

    def get_queryset(self):
        return Project.objects.with_counts().filter(
            organization=self.request.user.active_organization
        )

    def get_serializer_context(self):
        context = super(ProjectListAPI, self).get_serializer_context()
        context["created_by"] = self.request.user
        return context

    def perform_create(self, ser):
        try:
            project = ser.save(organization=self.request.user.active_organization)
        except IntegrityError as e:
            if str(e) == "UNIQUE constraint failed: project.title, project.created_by_id":
                raise ProjectExistException(
                    "Project with the same name already exists: {}".format(
                        ser.validated_data.get("title", "")
                    )
                )
            raise LabelStudioDatabaseException("Database error during project creation. Try again.")

    @swagger_auto_schema(tags=["Projects"])
    def get(self, request, *args, **kwargs):
        return super(ProjectListAPI, self).get(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Projects"], request_body=ProjectSerializer)
    def post(self, request, *args, **kwargs):
        return super(ProjectListAPI, self).post(request, *args, **kwargs)


class ProjectAPI(
    APIViewVirtualRedirectMixin, APIViewVirtualMethodMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    get:
    Get project by ID

    Retrieve information about a project by ID.

    patch:
    Update project

    Update project settings for a specific project.

    delete:
    Delete project

    Delete a project by specified project ID.
    """

    parser_classes = (JSONParser, FormParser, MultiPartParser)
    queryset = Project.objects.with_counts()
    permission_required = ViewClassPermission(
        GET=all_permissions.projects_view,
        DELETE=all_permissions.projects_delete,
        PATCH=all_permissions.projects_change,
        PUT=all_permissions.projects_change,
        POST=all_permissions.projects_create,
    )
    serializer_class = ProjectSerializer

    redirect_route = "projects:project-detail"
    redirect_kwarg = "pk"

    def get_queryset(self):
        return Project.objects.with_counts().filter(
            organization=self.request.user.active_organization
        )

    @swagger_auto_schema(tags=["Projects"])
    def get(self, request, *args, **kwargs):
        return super(ProjectAPI, self).get(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Projects"])
    def delete(self, request, *args, **kwargs):
        return super(ProjectAPI, self).delete(request, *args, **kwargs)

    @swagger_auto_schema(tags=["Projects"], request_body=ProjectSerializer)
    def patch(self, request, *args, **kwargs):
        project = self.get_object()
        label_config = self.request.data.get("label_config")

        # config changes can break view, so we need to reset them
        if label_config:
            try:
                has_changes = config_essential_data_has_changed(label_config, project.label_config)
            except KeyError:
                pass
            else:
                if has_changes:
                    View.objects.filter(project=project).all().delete()

        return super(ProjectAPI, self).patch(request, *args, **kwargs)

    def perform_destroy(self, instance):
        """Performance optimization for whole project deletion
        if we catch constraint error fallback to regular .delete() method"""
        try:
            task_annotation_qs = Annotation.objects.filter(task__project_id=instance.id)
            task_annotation_qs._raw_delete(task_annotation_qs.db)
            task_prediction_qs = Prediction.objects.filter(task__project_id=instance.id)
            task_prediction_qs._raw_delete(task_prediction_qs.db)
            task_locks_qs = TaskLock.objects.filter(task__project_id=instance.id)
            task_locks_qs._raw_delete(task_locks_qs.db)
            task_qs = Task.objects.filter(project_id=instance.id)
            task_qs._raw_delete(task_qs.db)
            instance.delete()
        except IntegrityError as e:
            logger.error("Fallback to cascade deleting after integrity_error: {}".format(str(e)))
            instance.delete()

    @swagger_auto_schema(auto_schema=None)
    def post(self, request, *args, **kwargs):
        return super(ProjectAPI, self).post(request, *args, **kwargs)

    @swagger_auto_schema(auto_schema=None)
    def put(self, request, *args, **kwargs):
        return super(ProjectAPI, self).put(request, *args, **kwargs)


class ProjectNextTaskAPI(generics.RetrieveAPIView):
    """get:
    Get next task to label

    Get the next task for labeling. If you enable Machine Learning in
    your project, the response might include a "predictions"
    field. It contains a machine learning prediction result for
    this task.

    """

    permission_required = all_permissions.tasks_view
    serializer_class = (
        TaskWithAnnotationsAndPredictionsAndDraftsSerializer  # using it for swagger API docs
    )

    def _get_random_unlocked(self, task_query, upper_limit=None):
        # get random task from task query, ignoring locked tasks
        n = task_query.count()
        if n > 0:
            upper_limit = upper_limit or n
            random_indices = np.random.permutation(upper_limit)
            task_query_only = task_query.only("overlap", "id")

            for i in random_indices:
                try:
                    task = task_query_only[int(i)]
                except IndexError as exc:
                    logger.error(
                        f"Task query out of range for {int(i)}, count={task_query_only.count()}. "
                        f"Reason: {exc}",
                        exc_info=True,
                    )
                except Exception as exc:
                    logger.error(exc, exc_info=True)
                else:
                    try:
                        task = Task.objects.select_for_update(skip_locked=True).get(pk=task.id)
                        if not task.has_lock():
                            return task
                    except Task.DoesNotExist:
                        logger.debug("Task with id {} locked".format(task.id))

    def _get_first_unlocked(self, tasks_query):
        # Skip tasks that are locked due to being taken by collaborators
        for task in tasks_query.all():
            try:
                task = Task.objects.select_for_update(skip_locked=True).get(pk=task.id)
                if not task.has_lock():
                    return task
            except Task.DoesNotExist:
                logger.debug("Task with id {} locked".format(task.id))

    def _try_ground_truth(self, tasks, project):
        """Returns task from ground truth set"""
        ground_truth = Annotation.objects.filter(task=OuterRef("pk"), ground_truth=True)
        not_solved_tasks_with_ground_truths = tasks.annotate(
            has_ground_truths=Exists(ground_truth)
        ).filter(has_ground_truths=True)
        if not_solved_tasks_with_ground_truths.exists():
            if project.sampling == project.SEQUENCE:
                return self._get_first_unlocked(not_solved_tasks_with_ground_truths)
            return self._get_random_unlocked(not_solved_tasks_with_ground_truths)

    def _try_tasks_with_overlap(self, tasks):
        """Filter out tasks without overlap (doesn't return next task)"""
        tasks_with_overlap = tasks.filter(overlap__gt=1)
        if tasks_with_overlap.exists():
            return None, tasks_with_overlap
        else:
            return None, tasks.filter(overlap=1)

    def _try_breadth_first(self, tasks):
        """Try to find tasks with maximum amount of annotations, since we are trying to label tasks as fast as possible"""

        # =======
        # This commented part is trying to solve breadth-first in a bit different way:
        # it selects first task where any amount of annotations have been already created
        # we've left it here to be able to select it through the project settings later
        # =======
        # annotations = Annotation.objects.filter(task=OuterRef('pk'), ground_truth=False)
        # not_solved_tasks_labeling_started = tasks.annotate(labeling_started=Exists(annotations))
        # not_solved_tasks_labeling_started_true = not_solved_tasks_labeling_started.filter(labeling_started=True)
        # if not_solved_tasks_labeling_started_true.exists():
        #     # try to complete tasks that are already in progress
        #     next_task = self._get_random(not_solved_tasks_labeling_started_true)
        #     return next_task

        tasks = tasks.annotate(annotations_count=Count("annotations"))
        max_annotations_count = tasks.aggregate(Max("annotations_count"))["annotations_count__max"]
        if max_annotations_count == 0:
            # there is no any labeled tasks found
            return

        # find any task with maximal amount of created annotations
        not_solved_tasks_labeling_started = tasks.annotate(
            reach_max_annotations_count=Case(
                When(annotations_count=max_annotations_count, then=Value(True)),
                default=Value(False),
                output_field=BooleanField(),
            )
        )
        not_solved_tasks_labeling_with_max_annotations = not_solved_tasks_labeling_started.filter(
            reach_max_annotations_count=True
        )
        if not_solved_tasks_labeling_with_max_annotations.exists():
            # try to complete tasks that are already in progress
            return self._get_random_unlocked(not_solved_tasks_labeling_with_max_annotations)

    def _try_uncertainty_sampling(self, tasks, project, user_solved_tasks_array):
        task_with_current_predictions = tasks.filter(
            predictions__model_version=project.model_version
        )
        if task_with_current_predictions.exists():
            logger.debug("Use uncertainty sampling")
            # collect all clusters already solved by user, count number of solved task in them
            user_solved_clusters = (
                project.prepared_tasks.filter(pk__in=user_solved_tasks_array)
                .annotate(cluster=Max("predictions__cluster"))
                .values_list("cluster", flat=True)
            )
            user_solved_clusters = Counter(user_solved_clusters)
            # order each task by the count of how many tasks solved in it's cluster
            cluster_num_solved_map = [
                When(predictions__cluster=k, then=v) for k, v in user_solved_clusters.items()
            ]

            num_tasks_with_current_predictions = (
                task_with_current_predictions.count()
            )  # WARNING! this call doesn't work after consequent annotate
            if cluster_num_solved_map:
                task_with_current_predictions = task_with_current_predictions.annotate(
                    cluster_num_solved=Case(
                        *cluster_num_solved_map, default=0, output_field=DecimalField()
                    )
                )
                # next task is chosen from least solved cluster and with lowest prediction score
                possible_next_tasks = task_with_current_predictions.order_by(
                    "cluster_num_solved", "predictions__score"
                )
            else:
                possible_next_tasks = task_with_current_predictions.order_by("predictions__score")

            num_annotators = project.annotators().count()
            if num_annotators > 1 and num_tasks_with_current_predictions > 0:
                # try to randomize tasks to avoid concurrent labeling between several annotators
                next_task = self._get_random_unlocked(
                    possible_next_tasks,
                    upper_limit=min(num_annotators + 1, num_tasks_with_current_predictions),
                )
            else:
                next_task = self._get_first_unlocked(possible_next_tasks)
        else:
            # uncertainty sampling fallback: choose by random sampling
            logger.debug(
                f"Uncertainty sampling fallbacks to random sampling "
                f"(current project.model_version={str(project.model_version)})"
            )
            next_task = self._get_random_unlocked(tasks)
        return next_task

    def _make_response(self, next_task, request, use_task_lock=True):
        """Once next task has chosen, this function triggers inference and prepare the API response"""
        user = request.user
        project = next_task.project

        if use_task_lock:
            # set lock for the task with TTL 3x time more then current average lead time (or 1 hour by default)
            next_task.set_lock(request.user)

        # call machine learning api and format response
        if project.show_collab_predictions and not next_task.predictions.exists():
            for ml_backend in project.ml_backends.all():
                ml_backend.predict_one_task(next_task)

        # serialize task
        context = {
            "request": request,
            "project": project,
            "resolve_uri": True,
            "proxy": bool_from_request(request.GET, "proxy", True),
        }
        serializer = TaskWithAnnotationsAndPredictionsAndDraftsSerializer(
            next_task, context=context
        )
        response = serializer.data

        annotations = []
        for c in response.get("annotations", []):
            if c.get("completed_by") == user.id and not (
                c.get("ground_truth") or c.get("honeypot")
            ):
                annotations.append(c)
        response["annotations"] = annotations

        # remove all predictions if we don't want to show it in the label stream
        if not project.show_collab_predictions:
            response["predictions"] = []

        return Response(response)

    @swagger_auto_schema(
        tags=["Projects"], responses={200: TaskWithAnnotationsAndPredictionsAndDraftsSerializer()}
    )
    def get(self, request, *args, **kwargs):
        project = get_object_with_check_and_log(request, Project, pk=self.kwargs["pk"])
        self.check_object_permissions(request, project)
        user = request.user

        # support actions api call from actions/next_task.py
        if hasattr(self, "prepared_tasks"):
            project.prepared_tasks = self.prepared_tasks
        # get prepared tasks from request params (filters, selected items)
        else:
            project.prepared_tasks = get_prepared_queryset(self.request, project)

        # detect solved and not solved tasks
        user_solved_tasks_array = (
            user.annotations.filter(ground_truth=False)
            .filter(Q(task__isnull=False))
            .values_list("task__pk", flat=True)
        )

        with conditional_atomic():
            not_solved_tasks = project.prepared_tasks.exclude(pk__in=user_solved_tasks_array)

            # if annotator is assigned for tasks, he must to solve it regardless of is_labeled=True
            assigned_flag = hasattr(self, "assignee_flag") and self.assignee_flag
            if not assigned_flag:
                not_solved_tasks = not_solved_tasks.annotate(
                    annotation_number=Count("annotations")
                ).filter(annotation_number__lte=project.maximum_annotations)

            not_solved_tasks_count = not_solved_tasks.count()

            # return nothing if there are no tasks remain
            if not_solved_tasks_count == 0:
                raise NotFound(f"There are no tasks remaining to be annotated by the user={user}")
            logger.debug(
                f"{not_solved_tasks_count} tasks that still need to be annotated for user={user}"
            )

            # ordered by data manager
            if assigned_flag:
                next_task = not_solved_tasks.first()
                if not next_task:
                    raise NotFound("No more tasks found")
                return self._make_response(next_task, request, use_task_lock=False)

            # If current user has already lock one task - return it (without setting the lock again)
            next_task = Task.get_locked_by(user, tasks=not_solved_tasks)
            if next_task:
                return self._make_response(next_task, request, use_task_lock=False)

            if project.show_ground_truth_first:
                logger.debug(
                    f"User={request.user} tries ground truth from {not_solved_tasks_count} tasks"
                )
                next_task = self._try_ground_truth(not_solved_tasks, project)
                if next_task:
                    return self._make_response(next_task, request)

            if project.show_overlap_first:
                # don't output anything - just filter tasks with overlap
                logger.debug(
                    f"User={request.user} tries overlap first from {not_solved_tasks_count} tasks"
                )
                _, not_solved_tasks = self._try_tasks_with_overlap(not_solved_tasks)

            # don't use this mode for data manager sorting, because the sorting becomes not obvious
            if project.sampling != project.SEQUENCE:
                # if there any tasks in progress (with maximum number of annotations), randomly sampling from them
                logger.debug(
                    f"User={request.user} tries depth first from {not_solved_tasks_count} tasks"
                )
                next_task = self._try_breadth_first(not_solved_tasks)
                if next_task:
                    return self._make_response(next_task, request)

            if project.sampling == project.UNCERTAINTY:
                logger.debug(
                    f"User={request.user} tries uncertainty sampling from {not_solved_tasks_count} tasks"
                )
                next_task = self._try_uncertainty_sampling(
                    not_solved_tasks, project, user_solved_tasks_array
                )

            elif project.sampling == project.UNIFORM:
                logger.debug(
                    f"User={request.user} tries random sampling from {not_solved_tasks_count} tasks"
                )
                next_task = self._get_random_unlocked(not_solved_tasks)

            elif project.sampling == project.SEQUENCE:
                logger.debug(
                    f"User={request.user} tries sequence sampling from {not_solved_tasks_count} tasks"
                )
                next_task = self._get_first_unlocked(not_solved_tasks)

            if next_task:
                return self._make_response(next_task, request)
            else:
                raise NotFound(
                    f"There are still some tasks to complete for the user={user}, but they seem to be locked by another user."
                )


class LabelConfigValidateAPI(generics.CreateAPIView):
    parser_classes = (JSONParser, FormParser, MultiPartParser)
    permission_classes = (AllowAny,)
    serializer_class = ProjectLabelConfigSerializer

    @swagger_auto_schema(
        tags=["Projects"],
        operation_summary="Validate label config",
        operation_description="Validate a labeling configuration for a project.",
        responses={200: "Validation success"},
    )
    def post(self, request, *args, **kwargs):
        return super(LabelConfigValidateAPI, self).post(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        try:
            serializer.is_valid(raise_exception=True)
        except RestValidationError as exc:
            context = self.get_exception_handler_context()
            response = exception_handler(exc, context)
            response = self.finalize_response(request, response)
            return response

        return Response(status=status.HTTP_204_NO_CONTENT)


class ProjectLabelConfigValidateAPI(generics.RetrieveAPIView):
    """Validate label config"""

    parser_classes = (JSONParser, FormParser, MultiPartParser)
    serializer_class = ProjectLabelConfigSerializer
    permission_required = all_permissions.projects_change
    queryset = Project.objects.all()

    @swagger_auto_schema(
        tags=["Projects"],
        operation_summary="Validate a label config",
        manual_parameters=[
            openapi.Parameter(
                name="label_config",
                type=openapi.TYPE_STRING,
                in_=openapi.IN_QUERY,
                description="labeling config",
            )
        ],
    )
    def post(self, request, *args, **kwargs):
        project = self.get_object()
        label_config = self.request.data.get("label_config")
        if not label_config:
            raise RestValidationError("Label config is not set or is empty")

        # check new config includes meaningful changes
        has_changed = config_essential_data_has_changed(label_config, project.label_config)
        project.validate_config(label_config)
        return Response(
            {"config_essential_data_has_changed": has_changed}, status=status.HTTP_200_OK
        )

    @swagger_auto_schema(auto_schema=None)
    def get(self, request, *args, **kwargs):
        return super(ProjectLabelConfigValidateAPI, self).get(request, *args, **kwargs)


class ProjectSummaryAPI(generics.RetrieveAPIView):
    parser_classes = (JSONParser,)
    serializer_class = ProjectSummarySerializer
    permission_required = all_permissions.projects_view
    queryset = ProjectSummary.objects.all()

    @swagger_auto_schema(tags=["Projects"], operation_summary="Project summary")
    def get(self, *args, **kwargs):
        return super(ProjectSummaryAPI, self).get(*args, **kwargs)


class TasksListAPI(
    generics.ListCreateAPIView,
    generics.DestroyAPIView,
    APIViewVirtualMethodMixin,
    APIViewVirtualRedirectMixin,
):
    """
    get:
    List project tasks

    Paginated list of tasks for a specific project.

    delete:
    Delete all tasks

    Delete all tasks from a specific project.
    """

    parser_classes = (JSONParser, FormParser)
    permission_required = ViewClassPermission(
        GET=all_permissions.tasks_view,
        POST=all_permissions.tasks_change,
        DELETE=all_permissions.tasks_delete,
    )
    serializer_class = TaskSerializer
    redirect_route = "projects:project-settings"
    redirect_kwarg = "pk"

    def get_queryset(self):
        project = generics.get_object_or_404(
            Project.objects.for_user(self.request.user), pk=self.kwargs.get("pk", 0)
        )
        tasks = Task.objects.filter(project=project)
        return paginator(tasks, self.request)

    @swagger_auto_schema(tags=["Projects"])
    def delete(self, request, *args, **kwargs):
        project = generics.get_object_or_404(
            Project.objects.for_user(self.request.user), pk=self.kwargs["pk"]
        )
        Task.objects.filter(project=project).delete()
        return Response(status=204)

    @swagger_auto_schema(**paginator_help("tasks", "Projects"))
    def get(self, *args, **kwargs):
        return super(TasksListAPI, self).get(*args, **kwargs)

    @swagger_auto_schema(auto_schema=None, tags=["Projects"])
    def post(self, *args, **kwargs):
        return super(TasksListAPI, self).post(*args, **kwargs)

    def get_serializer_context(self):
        context = super(TasksListAPI, self).get_serializer_context()
        context["project"] = get_object_with_check_and_log(
            self.request, Project, pk=self.kwargs["pk"]
        )
        return context

    def perform_create(self, serializer):
        project = get_object_with_check_and_log(self.request, Project, pk=self.kwargs["pk"])
        serializer.save(project=project)


class TemplateListAPI(generics.ListAPIView):
    parser_classes = (JSONParser, FormParser, MultiPartParser)
    permission_required = all_permissions.projects_view
    swagger_schema = None

    def list(self, request, *args, **kwargs):
        annotation_templates_dir = find_dir("annotation_templates")
        configs = []
        for config_file in pathlib.Path(annotation_templates_dir).glob("**/*.yml"):
            config = read_yaml(config_file)
            if config.get("image", "").startswith("/static") and settings.HOSTNAME:
                # if hostname set manually, create full image urls
                config["image"] = settings.HOSTNAME + config["image"]
            configs.append(config)
        template_groups_file = find_file(os.path.join("annotation_templates", "groups.txt"))
        with open(template_groups_file, encoding="utf-8") as f:
            groups = f.read().splitlines()
        logger.debug(f"{len(configs)} templates found.")
        return Response({"templates": configs, "groups": groups})


class ProjectSampleTask(generics.RetrieveAPIView):
    parser_classes = (JSONParser,)
    queryset = Project.objects.all()
    permission_required = all_permissions.projects_view
    serializer_class = ProjectSerializer
    swagger_schema = None

    def post(self, request, *args, **kwargs):
        label_config = self.request.data.get("label_config")
        if not label_config:
            raise RestValidationError("Label config is not set or is empty")

        project = self.get_object()
        return Response({"sample_task": project.get_sample_task(label_config)}, status=200)
